#line 26 "../noweb/g_smob.nw"
/* -*- geda-c -*-
 * gEDA - GPL Electronic Design Automation
 * libgeda - gEDA's library
 * Copyright (C) 1998-2000 Ales V. Hvezda
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA
 */


#line 11 "../noweb/g_smob.nw"
/* DO NOT read or edit this file ! Use ../noweb/g_smob.nw instead */

#line 50 "../noweb/g_smob.nw"
#include <config.h>

#include <math.h>
#include <stdio.h>
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <gtk/gtk.h>
#include <libguile.h>

#include "defines.h"
#include "struct.h"
#include "globals.h"

#include "../include/prototype.h"

#ifdef HAVE_LIBDMALLOC
#include <dmalloc.h>
#endif

#line 75 "../noweb/g_smob.nw"
static long attrib_smob_tag;

#line 86 "../noweb/g_smob.nw"
static scm_sizet g_free_attrib_smob(SCM attrib_smob)
{
  struct st_attrib_smob *attribute =
      (struct st_attrib_smob *) SCM_CDR(attrib_smob);
  scm_sizet size = sizeof(struct st_attrib_smob);

  free(attribute);
  return size;
}

#line 106 "../noweb/g_smob.nw"
static int
g_print_attrib_smob(SCM attrib_smob, SCM port, scm_print_state * pstate)
{
  struct st_attrib_smob *attribute =
      (struct st_attrib_smob *) SCM_CDR(attrib_smob);

  /* I don't think this is valid, because C does not support this sort
   * of evaluation, it will eval ALL the statements unlike pascal which
   * will stop after the first FALSE -Ales */
  if (attribute &&
      attribute->attribute &&
      attribute->attribute->object &&
      attribute->attribute->object->text &&
      attribute->attribute->object->text->string) {
    scm_puts("#<attribute ", port);
    scm_display(scm_makfrom0str
		(attribute->attribute->object->text->string), port);
    scm_puts(">", port);
  }

  /* non-zero means success */
  return 1;
}

#line 139 "../noweb/g_smob.nw"
/* Creates a name-value smob */
SCM g_make_attrib_smob(TOPLEVEL * curr_w, ATTRIB * curr_attr)
{
  struct st_attrib_smob *smob_attribute;

  smob_attribute =
      (struct st_attrib_smob *)
      scm_must_malloc(sizeof(struct st_attrib_smob), "attribute");

  smob_attribute->world = curr_w;
  smob_attribute->attribute = curr_attr;

  /* Assumes Guile version >= 1.3.2 */
  SCM_RETURN_NEWSMOB(attrib_smob_tag, smob_attribute);
}

#line 164 "../noweb/g_smob.nw"
SCM g_get_attrib_name_value(SCM attrib_smob)
{
  struct st_attrib_smob *attribute;
  char *name = NULL;
  char *value = NULL;
  SCM returned = SCM_EOL;


  SCM_ASSERT(SCM_NIMP(attrib_smob) &&
	     (SCM_CAR(attrib_smob) == attrib_smob_tag),
	     attrib_smob, SCM_ARG1, "get-attribute-name-value");

  attribute = (struct st_attrib_smob *) SCM_CDR(attrib_smob);

  if (attribute &&
      attribute->attribute &&
      attribute->attribute->object &&
      attribute->attribute->object->text->string) {
    o_attrib_get_name_value(attribute->attribute->object->text->string,
			    &name, &value);
    returned = scm_cons(scm_makfrom0str(name), scm_makfrom0str(value));
    if (name)
      free(name);
    if (value)
      free(value);
  }

  return returned;
}

#line 203 "../noweb/g_smob.nw"
SCM
g_set_attrib_value_internal(SCM attrib_smob, SCM scm_value,
			    TOPLEVEL ** world, OBJECT ** o_attrib,
			    char *new_string[])
{
  struct st_attrib_smob *attribute;
  char *name = NULL;
  char *value = NULL;
  char *old_value = NULL;

  SCM_ASSERT(SCM_NIMP(attrib_smob) &&
	     (SCM_CAR(attrib_smob) == attrib_smob_tag),
	     attrib_smob, SCM_ARG1, "set-attribute-value!");
  SCM_ASSERT(SCM_NIMP(scm_value) && SCM_STRINGP(scm_value),
	     scm_value, SCM_ARG2, "set-attribute-value!");

  attribute = (struct st_attrib_smob *) SCM_CDR(attrib_smob);
  value = SCM_STRING_CHARS(scm_value);

  if (attribute &&
      attribute->attribute &&
      attribute->attribute->object &&
      attribute->attribute->object->text &&
      attribute->attribute->object->text->string) {

    o_attrib_get_name_value(attribute->attribute->object->text->string,
			    &name, &old_value);

    *new_string = g_strconcat(name, "=", value, NULL);

    *world = attribute->world;
    *o_attrib = attribute->attribute->object;

    if (name)
      free(name);
    if (old_value)
      free(old_value);
  }

  return SCM_UNDEFINED;
}

#line 251 "../noweb/g_smob.nw"
void g_init_attrib_smob(void)
{

  attrib_smob_tag =
      scm_make_smob_type("attribute", sizeof(struct st_attrib_smob));
  scm_set_smob_mark(attrib_smob_tag, 0);
  scm_set_smob_free(attrib_smob_tag, g_free_attrib_smob);
  scm_set_smob_print(attrib_smob_tag, g_print_attrib_smob);

  scm_c_define_gsubr("get-attribute-name-value", 1, 0, 0,
		     g_get_attrib_name_value);

  return;
}
